@@ -3,16 +3,22 @@ import sys
 import traceback
 import typing
 from pathlib import Path
+from zipfile import ZipFile, ZIP_DEFLATED
 
-from PySide6.QtCore import QSettings, QSize, QSizeF, QObject, QRectF
+from PySide6.QtCore import QSettings, QSize, QSizeF, QObject, QRectF, QRect, QPoint, QBuffer
 from PySide6.QtGui import QFont, QPdfWriter, QPageSize, QPainter, QKeyEvent, \
     Qt, QCloseEvent, QPixmap, QColor, QTextDocument, QTextFormat, QTextCursor, \
-    QTextCharFormat, QPyTextObject
-from PySide6.QtWidgets import QApplication, QMainWindow, QMessageBox, QFileDialog, QInputDialog
+    QTextCharFormat, QPyTextObject, QImage
+from PySide6.QtSvg import QSvgGenerator
+from PySide6.QtWidgets import QApplication, QMainWindow, QMessageBox, QFileDialog, QInputDialog, QToolTip, \
+    QListWidgetItem
 
 import four_letter_blocks
+from four_letter_blocks.block_packer import BlockPacker
+from four_letter_blocks.clue_painter import CluePainter
 from four_letter_blocks.main_window import Ui_MainWindow
 from four_letter_blocks.puzzle import Puzzle
+from four_letter_blocks.puzzle_set import PuzzleSet
 
 DIAGRAM_TEXT_FORMAT = QTextFormat.UserObject + 1
 DIAGRAM_DATA = 1
@@ -33,10 +39,20 @@ class FourLetterBlocksWindow(QMainWindow):
         ui.save_action.triggered.connect(self.save)
         ui.save_as_action.triggered.connect(self.save_as)
         ui.export_action.triggered.connect(self.export)
+        ui.export_set_action.triggered.connect(self.export_laser)
 
         ui.shuffle_action.triggered.connect(self.shuffle)
         ui.options_action.triggered.connect(self.choose_font)
 
+        ui.crossword_files.currentRowChanged.connect(
+            self.select_crossword_file)
+        self.selected_crossword_file = -1
+        self.select_crossword_file(-1)
+        self.crossword_set: typing.Dict[str, Puzzle] = {}
+
+        ui.add_button.clicked.connect(self.add_crosswords)
+        ui.remove_button.clicked.connect(self.remove_crossword)
+
         sys.excepthook = self.on_error
         self.file_path: typing.Optional[Path] = None
         self.settings = get_settings()
@@ -50,12 +66,16 @@ class FourLetterBlocksWindow(QMainWindow):
         ui.blocks_text.focused.connect(self.blocks_changed)
         ui.clues_text.textChanged.connect(self.clues_changed)
 
-        self.state_fields = (self.ui.title_text,
-                             self.ui.grid_text,
-                             self.ui.clues_text,
-                             self.ui.blocks_text)
+        ui.warnings_label.setVisible(False)
+
+        self.state_fields = (ui.title_text,
+                             ui.grid_text,
+                             ui.clues_text,
+                             ui.blocks_text)
         self.clean_state = self.current_state = self.build_current_state()
         self.update_font()
+        ui.main_tabs.currentChanged.connect(self.select_tab)
+        self.select_tab(ui.main_tabs.currentIndex())
 
     def closeEvent(self, event: QCloseEvent):
         if self.can_abandon('quit'):
@@ -139,6 +159,68 @@ class FourLetterBlocksWindow(QMainWindow):
         self.ui.blocks_text.clear()
         self.record_clean_state()
 
+    def add_crosswords(self):
+        save_dir = self.get_save_dir()
+        kwargs = get_file_dialog_options()
+        file_names: typing.List[str]
+        file_names, selected_filter = QFileDialog.getOpenFileNames(
+            self,
+            'Open Crossword Files',
+            dir=save_dir,
+            filter='Text files (*.txt);;All files (*.*)',
+            **kwargs)
+        if not file_names:
+            return
+
+        crossword_files = self.ui.crossword_files
+        old_names = [((item := crossword_files.item(i)).text(), item.toolTip())
+                     for i in range(crossword_files.count())]
+        old_files = {file_name for text, file_name in old_names}
+
+        new_names = []
+        for file_name in file_names:
+            if file_name in old_files:
+                continue
+            with open(file_name) as puzzle_file:
+                puzzle = Puzzle.parse(puzzle_file)
+            self.crossword_set[file_name] = puzzle
+            new_names.append((puzzle.title + ' ' + puzzle.extras, file_name))
+        new_names.sort()
+
+        i = 0
+        while new_names:
+            new_name = new_names.pop(0)
+            while i < len(old_names):
+                old_name = old_names[i]
+                if new_name < old_name:
+                    old_names.insert(i, new_name)
+                    new_text, new_file = new_name
+                    new_item = QListWidgetItem(new_text)
+                    new_item.setToolTip(new_file)
+                    crossword_files.insertItem(i, new_item)
+                    i += 1
+                    break
+                i += 1
+            else:
+                old_names.append(new_name)
+                new_text, new_file = new_name
+                new_item = QListWidgetItem(new_text)
+                new_item.setToolTip(new_file)
+                crossword_files.addItem(new_item)
+                i += 1
+
+        self.summarize_crossword_set()
+
+    def summarize_crossword_set(self):
+        puzzles = []
+        crossword_files = self.ui.crossword_files
+        for i in range(crossword_files.count()):
+            file_name = crossword_files.item(i).toolTip()
+            puzzles.append(self.crossword_set[file_name])
+
+        puzzle_set = PuzzleSet(*puzzles)
+        self.statusBar().showMessage(puzzle_set.block_summary)
+
     def open(self):
         if not self.can_abandon('open a file'):
             return
@@ -160,31 +242,50 @@ class FourLetterBlocksWindow(QMainWindow):
         with file_path.open() as source_file:
             puzzle = Puzzle.parse(source_file)
         self.file_path = file_path
+
+        # Clear blocks first, to avoid extra warnings.
+        self.ui.blocks_text.setPlainText('')
+
         self.ui.title_text.setText(puzzle.title)
         self.ui.grid_text.setPlainText(puzzle.format_grid())
         self.ui.clues_text.setPlainText(puzzle.format_clues())
         self.ui.blocks_text.setPlainText(puzzle.format_blocks())
         self.record_clean_state()
 
+    def remove_crossword(self):
+        if self.selected_crossword_file >= 0:
+            item = self.ui.crossword_files.takeItem(
+                self.selected_crossword_file)
+            file_name = item.toolTip()
+            del self.crossword_set[file_name]
+            self.selected_crossword_file = self.ui.crossword_files.currentRow()
+            self.summarize_crossword_set()
+
     def record_clean_state(self):
         self.clean_state = self.build_current_state()
         self.is_state_changed()  # Update dirty display.
 
     def save_as(self):
+        file_name = self.get_save_file_name(
+            'Save puzzle',
+            'Text files (*.txt);;All files (*.*)')
+        if not file_name:
+            return
+
+        self.file_path = Path(file_name)
+        self.save()
+
+    def get_save_file_name(self, caption, file_filter):
         save_dir = self.get_save_dir()
         kwargs = get_file_dialog_options()
         file_name, _ = QFileDialog.getSaveFileName(
             self,
-            'Save puzzle',
+            caption,
             dir=save_dir,
-            filter='Text files (*.txt);;All files (*.*)',
+            filter=file_filter,
             **kwargs)
-        if not file_name:
-            return
-        self.file_path = Path(file_name)
-        self.settings.setValue('save_path', str(self.file_path))
-
-        self.save()
+        self.settings.setValue('save_path', file_name)
+        return file_name
 
     def get_save_dir(self):
         save_path = self.settings.value('save_path')
@@ -227,12 +328,109 @@ class FourLetterBlocksWindow(QMainWindow):
             return
         self.settings.setValue('save_path', file_name)
         file_path = Path(file_name)
-        if file_path.suffix.lower() == '.pdf':
+        file_suffix = file_path.suffix.lower()
+        if file_suffix == '.pdf':
             self.export_pdf(file_path)
-        elif file_path.suffix.lower() == '.png':
+        elif file_suffix == '.png':
             self.export_png(file_path)
         else:
             self.export_md(file_path)
+        self.statusBar().showMessage(f'Exported to {file_path.name}.')
+
+    @staticmethod
+    def is_field_filled(field, label, button):
+        if field.text():
+            return True
+        message = f"Select a {label.text().lower()} before exporting."
+        point = button.mapToGlobal(QPoint(0, 0))
+        QToolTip.showText(point, message)
+        return False
+
+    def export_laser(self):
+        ui = self.ui
+        if ui.crossword_files.count() < 2:
+            point = ui.add_button.mapToGlobal(QPoint(0, 0))
+            QToolTip.showText(point, "Add more crosswords before exporting.")
+            return
+        save_dir = self.get_save_dir()
+        kwargs = get_file_dialog_options()
+        file_name, _ = QFileDialog.getSaveFileName(
+            self,
+            'Export puzzle set',
+            dir=save_dir,
+            filter=';;'.join(('Zip files (*.zip)',
+                              'All files (*.*)')),
+            **kwargs)
+        if not file_name:
+            return
+        self.settings.setValue('save_path', file_name)
+
+        packer = BlockPacker(16, 20, tries=10_000)
+        puzzles = list(self.crossword_set.values())
+        puzzles.sort(key=lambda p: (p.grid.width, p.title))
+        puzzle_set = PuzzleSet(*puzzles, block_packer=packer)
+
+        svg_buffer = QBuffer()
+        generator = QSvgGenerator()
+        generator.setOutputDevice(svg_buffer)
+        generator.setSize(QSize(8250, 10500))
+        generator.setResolution(1000)  # dots per inch
+        generator.setViewBox(QRect(0, 0, 8250, 10500))
+
+        painter = QPainter(generator)
+        puzzle_set.square_size = generator.width() / 17
+        nick_radius = 5  # DPI is 1000
+        puzzle_set.draw_cuts(painter, nick_radius)
+        painter.end()
+
+        front_buffer = QBuffer()
+        front_image = QImage(2475, 3150, QImage.Format_RGB32)
+        painter = QPainter(front_image)
+        painter.fillRect(front_image.rect(), 'white')
+        puzzle_set.square_size = front_image.width() / 17
+        puzzle_set.draw_front(painter)
+        painter.end()
+        success = front_image.save(front_buffer, 'PNG')
+        assert success
+
+        back_buffer = QBuffer()
+        back_image = QImage(2475, 3150, QImage.Format_RGB32)
+        painter = QPainter(back_image)
+        painter.fillRect(back_image.rect(), 'white')
+        puzzle_set.draw_back(painter)
+        painter.end()
+        success = back_image.save(back_buffer, 'PNG')
+        assert success
+
+        """ Booklet page images are 1575 x 2475. Safety margin is 75 pixels on
+        every side. """
+        page_buffers = []
+        clue_painter = CluePainter(
+            *puzzles,
+            font_size=60,
+            margin=75,
+            intro_text='Solve each crossword puzzle with the pieces that match '
+                       'the colour of the clue numbers. Good luck!\n',
+            footer_text='https://donkirkby.github.io/four-letter-blocks')
+        page_image = QImage(1575, 2475, QImage.Format_RGB32)
+        while not clue_painter.is_finished:
+            painter = QPainter(page_image)
+            painter.fillRect(page_image.rect(), 'white')
+            clue_painter.draw_page(painter)
+            painter.end()
+            page_buffer = QBuffer()
+            success = page_image.save(page_buffer, 'PNG')
+            assert success
+            page_buffers.append(page_buffer)
+
+        with ZipFile(file_name, 'w', compression=ZIP_DEFLATED) as zip_file:
+            zip_file.writestr('cuts.svg', svg_buffer.data())
+            zip_file.writestr('front.png', front_buffer.data())
+            zip_file.writestr('back.png', back_buffer.data())
+            for page_number, page_buffer in enumerate(page_buffers, 1):
+                zip_file.writestr(f'page{page_number}.png', page_buffer.data())
+
+        self.statusBar().showMessage(f'Exported to {file_name}.')
 
     def export_pdf(self, file_path: Path):
         file_name = str(file_path)
@@ -266,8 +464,6 @@ class FourLetterBlocksWindow(QMainWindow):
 
         document.print_(pdf)
 
-        self.statusBar().showMessage(f'Exported to {file_path.name}.')
-
     def export_png(self, file_path: Path):
         puzzle = self.parse_puzzle()
         width, height = 640, 2000
@@ -288,13 +484,13 @@ class FourLetterBlocksWindow(QMainWindow):
             print(file=file)
             print('Across  ', file=file)
             for clue in puzzle.across_clues:
-                formatted_clue = f'**{clue.format_number()}.** {clue.text}  '
-                print(formatted_clue, file=file)
+                print(f'**{clue.format_number()}.** {clue.format_text()}  ',
+                      file=file)
             print(file=file)
             print('Down  ', file=file)
             for clue in puzzle.down_clues:
-                formatted_clue = f'**{clue.format_number()}.** {clue.text}  '
-                print(formatted_clue, file=file)
+                print(f'**{clue.format_number()}.** {clue.format_text()}  ',
+                      file=file)
 
     def parse_puzzle(self):
         puzzle = Puzzle.parse_sections(self.ui.title_text.text(),
@@ -324,9 +520,14 @@ class FourLetterBlocksWindow(QMainWindow):
         if not self.is_state_changed():
             return
         puzzle = self.parse_puzzle()
-        block_summary = puzzle.display_block_sizes()
+        warnings = puzzle.check_style()
+        self.ui.warnings_label.setVisible(bool(warnings))
+        if warnings:
+            warnings.insert(0, 'Warnings')
+            self.ui.warnings_label.setText('\n  '.join(warnings))
+        block_summary = puzzle.display_block_summary()
         if block_summary:
-            self.statusBar().showMessage(f'Block sizes: {block_summary}.')
+            self.statusBar().showMessage(block_summary)
 
     def clues_changed(self):
         if not self.is_state_changed():
@@ -347,6 +548,26 @@ class FourLetterBlocksWindow(QMainWindow):
             self.settings.setValue('font_size', font_size)
             self.update_font()
 
+    def select_tab(self, tab_index):
+        is_single = tab_index == 0
+        self.ui.new_action.setEnabled(is_single)
+        self.ui.open_action.setEnabled(is_single)
+        self.ui.save_action.setEnabled(is_single)
+        self.ui.save_as_action.setEnabled(is_single)
+        self.ui.shuffle_action.setEnabled(is_single)
+        self.ui.export_action.setEnabled(is_single)
+        self.ui.export_set_action.setEnabled(not is_single)
+        if is_single:
+            puzzle = self.parse_puzzle()
+            block_summary = puzzle.display_block_summary()
+            self.statusBar().showMessage(block_summary)
+        else:
+            self.summarize_crossword_set()
+
+    def select_crossword_file(self, file_index):
+        self.selected_crossword_file = file_index
+        self.ui.remove_button.setEnabled(file_index >= 0)
+
     def update_font(self):
         font_size = self.settings.value('font_size', 11, int)
         font = self.font()
@@ -365,9 +586,12 @@ class FourLetterBlocksWindow(QMainWindow):
 
 
 class BlockDiagram(QPyTextObject):
-    def __init__(self, puzzle: Puzzle, parent: QObject = None):
+    def __init__(self, puzzle: Puzzle,
+                 position_map: typing.Dict = None,
+                 parent: QObject = None):
         super().__init__(parent)
         self.puzzle = puzzle
+        self.position_map = position_map
 
     # noinspection PyPep8Naming,PyShadowingBuiltins
     def intrinsicSize(self,
