@@ -1,13 +1,14 @@
 import re
 import typing
-from collections import Counter
+from collections import Counter, defaultdict
 from dataclasses import dataclass, field
 from html import escape
 from itertools import chain
 from operator import attrgetter
 from random import shuffle
+from textwrap import dedent
 
-from PySide6.QtGui import QPainter, QTextDocument
+from PySide6.QtGui import QPainter, QTextDocument, QTextCursor
 
 from four_letter_blocks.clue import Clue
 from four_letter_blocks.grid import Grid
@@ -16,8 +17,8 @@ from four_letter_blocks.block import Block
 
 @dataclass
 class Puzzle:
-    HINT = 'Clue numbers are shuffled: 1 Across might not be in the top left.'
-    SUIT_HINT = "Each corner uses a different suit."
+    HINT = 'Clue numbers are shuffled: 1 Across might not be the top left.'
+    SUIT_HINT = "Each corner has its own suit."
     DEFAULT_ROW_LENGTH = 16  # Number of squares across a diagram.
 
     title: str
@@ -27,6 +28,7 @@ class Puzzle:
     across_clues: typing.List[Clue] = field(init=False)
     down_clues: typing.List[Clue] = field(init=False)
     use_suits: bool = False
+    use_text: bool = True
 
     @staticmethod
     def parse(source_file: typing.IO) -> 'Puzzle':
@@ -123,7 +125,7 @@ class Puzzle:
                     except KeyError:
                         continue
                     clue_chars[match.start():match.end()] = reference
-                clue.text = ''.join(clue_chars)
+                clue.text_with_reference = ''.join(clue_chars)
         clue_key = attrgetter('suit', 'number')
         self.across_clues.sort(key=clue_key)
         self.down_clues.sort(key=clue_key)
@@ -150,6 +152,23 @@ class Puzzle:
                 square.x *= ratio
                 square.y *= ratio
 
+    @property
+    def extras(self):
+        messages = []
+        pairs = 'JLSZ'
+        shape_counts = self.shape_counts
+        for shape, count in sorted(shape_counts.items()):
+            pair_index = pairs.find(shape)
+            if pair_index < 0:
+                if count % 2 != 0:
+                    messages.append(f'{shape}+')
+            else:
+                mirror = pairs[pair_index+1 - 2*(pair_index % 2)]
+                mirror_count = shape_counts[mirror]
+                if count > mirror_count:
+                    messages.append(f'{shape}+{count-mirror_count}')
+        return ', '.join(messages)
+
     def draw_blocks(self,
                     painter: QPainter,
                     square_size: int = None,
@@ -172,7 +191,7 @@ class Puzzle:
         row_count = 0
         x_limit = painter.window().width() - self.square_size
         is_active_row = False
-        for block in self.blocks:
+        for block in self.sorted_blocks():
             if x_limit < x + block.width - x_start:
                 x = self.square_size
                 if is_active_row:
@@ -183,7 +202,7 @@ class Puzzle:
             if is_active_row:
                 block.x = x
                 block.y = y
-                block.draw(painter)
+                block.draw(painter, use_text=self.use_text)
             line_height = max(line_height, block.height)
             x += block.width + gap
         if row_index is None:
@@ -192,6 +211,10 @@ class Puzzle:
             y += line_height
         return y
 
+    def sorted_blocks(self):
+        """ Blocks sorted by increasing height. """
+        return sorted(self.blocks, key=lambda b: b.height)
+
     def row_heights(self,
                     window_width: int = None,
                     square_size: int = None) -> typing.List[int]:
@@ -209,7 +232,7 @@ class Puzzle:
         x = self.square_size
         line_height = 0
         x_limit = window_width - square_size
-        for block in self.blocks:
+        for block in self.sorted_blocks():
             if x_limit < x + block.width:
                 x = self.square_size
                 row_heights.append(line_height + gap)
@@ -254,6 +277,103 @@ class Puzzle:
                 rows[y][x] = block_letter
         return '\n'.join(''.join(row) for row in rows)
 
+    def display_block_summary(self) -> str:
+        sections = []
+        block_sizes = self.display_block_sizes()
+        sections.append(f'Block sizes: {block_sizes}')
+        shape_count_text = ', '.join(
+            f'{shape}: {count}'
+            for shape, count in sorted(self.shape_counts.items()))
+        if shape_count_text:
+            sections.append('Shapes: ' + shape_count_text)
+        return ', '.join(sections)
+
+    def check_style(self) -> typing.List[str]:
+        warnings = list(self.check_symmetry())
+        warnings.extend(self.check_repeats())
+        warnings.extend(self.check_word_length())
+        return warnings
+
+    def check_word_length(self):
+        short_warnings = []
+        complete_warnings = []
+        for block in self.blocks:
+            for square1 in block.squares:
+                for word, dx, dy in ((square1.across_word, 1, 0),
+                                     (square1.down_word, 0, 1)):
+                    if word is None:
+                        continue
+                    x1 = square1.x
+                    y1 = square1.y
+                    word_length = len(word)
+                    start = (x1 + 1, y1 + 1)
+                    end = (x1 + word_length * dx + dy, y1 + word_length * dy + dx)
+                    if word_length == 2:
+                        short_warnings.append(
+                            (start,
+                             end,
+                             f'two-letter word at {start} and {end}'))
+                    if block.marker == Block.UNUSED:
+                        continue
+                    block_coordinates = block.calculate_coordinates()
+                    for i in range(1, word_length):
+                        square2 = self.grid[x1 + i * dx, y1 + i * dy]
+                        if (square2.x, square2.y) not in block_coordinates:
+                            break
+                    else:
+                        complete_warnings.append((
+                            start,
+                            end,
+                            f'complete word on one block from {start} to {end}'))
+        short_warnings.sort()
+        yield from (warning for start, end, warning in short_warnings)
+        complete_warnings.sort()
+        yield from (warning for start, end, warning in complete_warnings)
+
+    def check_repeats(self):
+        word_counts = Counter()
+        for x in range(self.grid.width):
+            for y in range(self.grid.height):
+                square = self.grid[x, y]
+                if square is not None:
+                    word_counts[square.across_word] += 1
+                    word_counts[square.down_word] += 1
+        word_counts[None] = 0
+        repeats = [word for word, count in word_counts.items() if count > 1]
+        repeats.sort()
+        yield from (f'repeated word {word}' for word in repeats)
+
+    def check_symmetry(self):
+        grid = self.grid
+        if grid.width != grid.height:
+            yield 'not square'
+            return
+        x_limit = (grid.width + 1) // 2
+        for x1 in range(x_limit):
+            y_limit = grid.height if x1*2 < grid.width-1 else (grid.height+1)//2
+            for y1 in range(y_limit):
+                is_block1 = grid[x1, y1] is None
+                x2 = grid.width - x1 - 1
+                y2 = grid.height - y1 - 1
+                is_block2 = grid[x2, y2] is None
+                if is_block1 != is_block2:
+                    yield (f'symmetry broken at {(x1 + 1, y1 + 1)} and '
+                           f'{(x2 + 1, y2 + 1)}')
+
+    @property
+    def shape_counts(self):
+        return Counter(block.shape
+                       for block in self.blocks
+                       if block.shape is not None)
+
+    @property
+    def shape_blocks(self):
+        d = defaultdict(list)
+        for block in self.blocks:
+            if block.shape is not None:
+                d[block.shape].append(block)
+        return d
+
     def display_block_sizes(self) -> str:
         correct_markers = {block.marker
                            for block in self.blocks
@@ -277,7 +397,19 @@ class Puzzle:
         shuffle(self.blocks)
         self.number_clues()
 
-    def build_clues(self, document: QTextDocument):
+    @property
+    def face_colour(self):
+        return self.blocks[0].face_colour
+
+    @face_colour.setter
+    def face_colour(self, face_colour):
+        for block in self.blocks:
+            block.face_colour = face_colour
+
+    def build_clues(self, document: QTextDocument, show_link=True):
+        cursor = QTextCursor(document)
+        cursor.movePosition(cursor.End)
+
         font_size = document.defaultFont().pixelSize()
         padding = font_size // 5
         document.setDefaultStyleSheet(f"""\
@@ -285,35 +417,41 @@ h1 {{text-align: center}}
 hr.footer {{line-height:10px}}
 p.footer {{page-break-after: always}}
 td {{padding: {padding} }}
-td.num {{text-align: right}}
+td.num {{text-align: right; background: {self.face_colour.name()}}}
 a {{color: black}}
 """)
         across_table = build_clue_table(self.across_clues)
         down_table = build_clue_table(self.down_clues)
         hints = self.build_hints()
-        document.setHtml(f"""\
-<h1>{escape(self.title)}</h1>
-<p>{escape(hints)}</p>
-<hr>
-<table>
-<tr><td>Across</td><td>Down</td></tr>
-<tr><td>
-{across_table}
-</td><td>
-{down_table}
-</td></tr>
-</table>
-<hr class="footer">
-<p class="footer"><center><a href="https://donkirkby.github.io/four-letter-blocks"
->https://donkirkby.github.io/four-letter-blocks</a></center></p>
-<p></p>
-""")
+        html = dedent(f"""\
+            <h1>{escape(self.title)}</h1>
+            <p>{escape(hints)}</p>
+            <hr>
+            <table>
+            <tr><td>Across</td><td>Down</td></tr>
+            <tr><td>
+            {across_table}
+            </td><td>
+            {down_table}
+            </td></tr>
+            </table>
+        """)
+        if show_link:
+            html += dedent("""\
+                <hr class="footer">
+                <p class="footer"><center><a href="https://donkirkby.github.io/four-letter-blocks"
+                >https://donkirkby.github.io/four-letter-blocks</a></center></p>
+                <p></p>
+            """)
+        cursor.insertHtml(html)
+        cursor.movePosition(cursor.End)
 
     def build_hints(self):
         hints = self.HINT
         if self.use_suits:
             hints += ' '
             hints += self.SUIT_HINT
+        hints += f' {len(self.blocks)} pieces.'
         return hints
 
 
@@ -321,7 +459,7 @@ def build_clue_table(clues: typing.Sequence[Clue]) -> str:
     rows = []
     for clue in clues:
         rows.append(f'<tr><td class="num">{clue.format_number()}.</td>'
-                    f'<td>{escape(clue.text)}</td></tr>')
+                    f'<td>{escape(clue.format_text())}</td></tr>')
     return f'<table>{"".join(rows)}</table>'
 
 
